"""
Metrc API URLs | Cannlytics Module
Copyright (c) 2021-2022 Cannlytics and Cannlytics Contributors

Authors: Keegan Skeate <https://github.com/keeganskeate>
Created: 11/5/2021
Updated: 11/5/2021
License: <https://github.com/cannlytics/cannlytics/blob/main/LICENSE>
"""

METRC_API_BASE_URL = 'https://api-%s.metrc.com'
METRC_API_BASE_URL_TEST = 'https://sandbox-api-%s.metrc.com'
METRC_BATCHES_URL = '/plantbatches/v1/%s'
METRC_EMPLOYEES_URL = '/employees/v1/'
METRC_FACILITIES_URL = '/facilities/v1/'
METRC_LOCATIONS_URL = '/locations/v1/%s'
METRC_HARVESTS_URL = '/harvests/v1/%s'
METRC_ITEMS_URL = '/items/v1/%s'
METRC_LAB_RESULTS_URL = '/labtests/v1/%s'
METRC_PACKAGES_URL = '/packages/v1/%s'
METRC_PATIENTS_URL = '/patients/v1/%s'
METRC_PLANTS_URL = '/plants/v1/%s'
METRC_RECEIPTS_URL = '/sales/v1/receipts/%s'
METRC_SALES_URL = '/sales/v1/%s'
METRC_STRAINS_URL = '/strains/v1/%s'
METRC_TRANSACTIONS_URL = '/sales/v1/transactions/%s'
METRC_TRANSFERS_URL = '/transfers/v1/%s'
METRC_TRANSFER_PACKAGES_URL = '/transfers/v1/delivery/%s/%s'
METRC_TRANSFER_TEMPLATE_URL = '/transfers/v1/templates/%s'
METRC_UOM_URL = '/unitsofmeasure/v1/active'
