"""
Cannlytics CoA Data Initialization | Cannlytics
Copyright (c) 2022 Cannlytics

Authors: Keegan Skeate <https://github.com/keeganskeate>
Created: 7/21/2022
Updated: 8/30/2022
"""
try:
    from .coas import CoADoc
    __all__ = [
        CoADoc,
    ]
except ImportError:
    pass
