"""
Sample Management | Cannlytics

Authors: Keegan Skeate <https://github.com/keeganskeate>  
Created: 6/23/2021  
Updated: 6/23/2021  
License: MIT License <https://opensource.org/licenses/MIT>  

Manage your laboratory samples.
"""

def get_sample_pictures():
    """Get sample pictures taken and stored on a lab computer."""
    # pics = []
    # return pics
    return NotImplementedError


if __name__ == '__main__':
    
    # TODO: Test get sample pictures.
    get_sample_pictures()

