"""
Project Management | Cannlytics

Authors: Keegan Skeate <https://github.com/keeganskeate>  
Created: 11/2/2021  
Updated: 11/2/2021  
License: MIT License <https://opensource.org/licenses/MIT>  

Manage your laboratory projects.
"""

# TODO: Implement various project helper functions here.
