"""
Measurement Management | Cannlytics

Authors: Keegan Skeate <https://github.com/keeganskeate>  
Created: 11/2/2021  
Updated: 11/2/2021  
License: MIT License <https://opensource.org/licenses/MIT>  

Manage your laboratory measurements.
"""

def record_measurement(config):
    """Record a laboratory measurement, e.g. a measured sample weight."""
    # pics = []
    # return pics
    return NotImplementedError


if __name__ == '__main__':
    
    # TODO: Test recording a measurement.
    record_measurement({})

