"""
Statistics | Cannlytics

Authors: Keegan Skeate <https://github.com/keeganskeate>  
Created: 6/23/2021  
Updated: 6/23/2021  
License: MIT License <https://opensource.org/licenses/MIT>  

Calculate statistics on your laboratory performance.
"""


def calculate_turnaround_time():
    """Calculate the average turnaround time of a group of samples or
    projects."""
    return NotImplementedError


def calculate_samples_tested():
    """Calculate the total number of samples tested."""
    return NotImplementedError


def calculate_organizations_served():
    """Calculate the total number of organizations served."""
    return NotImplementedError


def calculate_projects_completed():
    """Calculate the total number of projects completed."""
    return NotImplementedError


def calculate_avg_failure_rate():
    """Calculate the avergage failure rate for an analyte in a group of samples."""
    return NotImplementedError


def calculate_avg_result():
    """Calculate the avergage result for an analyte in a group of samples."""
    return NotImplementedError
